#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
python-cfconfigurator
"""

__program__ = "cfconfigurator"
__version__ = "0.2.1"
__author__ = "Jose Riguera"
__year__ = "2016"
__email__ = "<jose.riguera@springer.com>"
__license__ = "MIT"
